from __future__ import annotations

from collections import OrderedDict
from typing import Any, Dict, Iterable, Iterator, Mapping, Optional, Type, TypeVar

__version__ = "2.2.5"

_K = TypeVar("_K")
_V = TypeVar("_V")


class immutabledict(Mapping[_K, _V]):
    """
    An immutable wrapper around dictionaries that implements
    the complete :py:class:`collections.Mapping` interface.
    It can be used as a drop-in replacement for dictionaries
    where immutability is desired.
    """

    dict_cls: Type[Dict[Any, Any]] = dict

    @classmethod
    def fromkeys(
        cls, seq: Iterable[_K], value: Optional[_V] = None
    ) -> "immutabledict[_K, _V]":
        return cls(cls.dict_cls.fromkeys(seq, value))

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._dict = self.dict_cls(*args, **kwargs)
        self._hash: Optional[int] = None

    def __getitem__(self, key: _K) -> _V:
        return self._dict[key]

    def __contains__(self, key: object) -> bool:
        return key in self._dict

    def copy(self, **add_or_replace: _V) -> immutabledict[_K, _V]:
        return self.__class__(self, **add_or_replace)

    def __iter__(self) -> Iterator[_K]:
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def __repr__(self) -> str:
        return "%s(%r)" % (self.__class__.__name__, self._dict)

    def __hash__(self) -> int:
        if self._hash is None:
            h = 0
            for key, value in self.items():
                h ^= hash((key, value))
            self._hash = h

        return self._hash

    def __or__(self, other: Any) -> immutabledict[_K, _V]:
        if not isinstance(other, (dict, self.__class__)):
            return NotImplemented
        new = dict(self)
        new.update(other)
        return self.__class__(new)

    def __ror__(self, other: Any) -> Dict[Any, Any]:
        if not isinstance(other, (dict, self.__class__)):
            return NotImplemented
        new = dict(other)
        new.update(self)
        return new

    def __ior__(self, other: Any) -> immutabledict[_K, _V]:
        raise TypeError(f"'{self.__class__.__name__}' object is not mutable")


class ImmutableOrderedDict(immutabledict):
    """
    An immutabledict subclass that maintains key order.

    Not necessary anymore as for Python >= 3.7 order is guaranteed with the normal
    immutabledict class, but kept for compatibility purpose.
    """

    dict_cls = OrderedDict



import pytest

from immutabledict import ImmutableOrderedDict, immutabledict


class TestImmutableDict:
    def test_cannot_assign_value(self):
        with pytest.raises(AttributeError):
            immutabledict().setitem("key", "value")

    def test_from_keys(self):
        keys = ("a", "b", "c")
        immutable_dict = immutabledict.fromkeys(keys)
        assert "a" in immutable_dict
        assert "b" in immutable_dict
        assert "c" in immutable_dict

    def test_init_and_compare(self):
        normal_dict = {"a": "value", "b": "other_value"}
        immutable_dict = immutabledict(normal_dict)
        assert immutable_dict == normal_dict

    def test_get_existing(self):
        immutable_dict = immutabledict({"a": "value"})
        assert immutable_dict["a"] == "value"

    def test_get_not_existing(self):
        immutable_dict = immutabledict({"a": "value"})
        with pytest.raises(KeyError):
            immutable_dict["b"]

    def test_contains(self):
        immutable_dict = immutabledict({"a": "value"})
        assert "a" in immutable_dict

    def test_contains_not_existing(self):
        immutable_dict = immutabledict({"a": "value"})
        assert "b" not in immutable_dict

    def test_copy(self):
        original = immutabledict({"a": "value"})
        copy = original.copy()
        assert original == copy

    def test_iter(self):
        immutable_dict = immutabledict({"a": "value", "b": "other_value"})
        itered_keys = {x for x in immutable_dict}
        assert immutable_dict.keys() == itered_keys

    def test_len(self):
        immutable_dict = immutabledict({"a": "value", "b": "other_value"})
        assert len(immutable_dict) == 2

    def test_len_empty(self):
        immutable_dict = immutabledict({})
        assert len(immutable_dict) == 0

    def test_repr(self):
        immutable_dict = immutabledict({"a": "value", "b": "other_value"})
        repr_ret = repr(immutable_dict)
        assert repr_ret.startswith("immutabledict")
        assert repr_ret.endswith(")")

    def test_repr_should_eval(self):
        immutable_dict = immutabledict({"a": "value", "b": "other_value"})
        eval_ret = eval(repr(immutable_dict))
        assert immutable_dict == eval_ret

    def test_hash(self):
        first_dict = immutabledict({"a": "value", "b": "other_value"})
        second_dict = immutabledict({"a": "value", "b": "other_value"})

        assert hash(first_dict) == hash(second_dict)

    def test_union_operator_merge(self):
        first_dict = immutabledict({"a": "a", "b": "b"})
        second_dict = immutabledict({"a": "A", "c": "c"})
        merged_dict = first_dict | second_dict
        assert isinstance(merged_dict, immutabledict)
        assert merged_dict == {
            "a": "A",
            "b": "b",
            "c": "c",
        }
        assert first_dict == {"a": "a", "b": "b"}
        assert second_dict == {"a": "A", "c": "c"}

    def test_union_operator_merge_with_dict(self):
        first_dict = dict({"a": "a", "b": "b"})
        second_dict = immutabledict({"a": "A", "c": "c"})
        merged_dict = first_dict | second_dict
        assert isinstance(merged_dict, dict)
        assert merged_dict == {
            "a": "A",
            "b": "b",
            "c": "c",
        }
        assert first_dict == {"a": "a", "b": "b"}
        assert second_dict == {"a": "A", "c": "c"}

        first_dict = immutabledict({"a": "a", "b": "b"})
        second_dict = dict({"a": "A", "c": "c"})
        merged_dict = first_dict | second_dict
        assert isinstance(merged_dict, immutabledict)
        assert merged_dict == {
            "a": "A",
            "b": "b",
            "c": "c",
        }
        assert first_dict == {"a": "a", "b": "b"}
        assert second_dict == {"a": "A", "c": "c"}

    def test_union_operator_merge_fail(self):
        first_dict = immutabledict({"a": "a", "b": "b"})

        with pytest.raises(TypeError):
            first_dict | 0

        with pytest.raises(TypeError):
            0 | first_dict

    def test_union_operator_update(self):
        first_dict = immutabledict({"a": "a", "b": "b"})
        second_dict = immutabledict({"a": "A", "c": "c"})

        with pytest.raises(TypeError):
            first_dict |= second_dict

    def test_union_operator_update_with_dict(self):
        first_dict = dict({"a": "a", "b": "b"})
        second_dict = immutabledict({"a": "A", "c": "c"})

        first_dict |= second_dict
        assert isinstance(first_dict, dict)
        assert first_dict == {
            "a": "A",
            "b": "b",
            "c": "c",
        }
        assert second_dict == {"a": "A", "c": "c"}

        first_dict = immutabledict({"a": "a", "b": "b"})
        second_dict = dict({"a": "A", "c": "c"})

        with pytest.raises(TypeError):
            first_dict |= second_dict
        assert isinstance(first_dict, immutabledict)
        assert first_dict == {"a": "a", "b": "b"}
        assert second_dict == {"a": "A", "c": "c"}


class TestImmutableOrderedDict:
    def test_ordered(self):
        ordered = ImmutableOrderedDict(
            {
                "a": "1",
                "b": "2",
                "c": "3",
            }
        )
        itered_keys = [x for x in ordered]
        assert itered_keys[0] == "a"
        assert itered_keys[1] == "b"
        assert itered_keys[2] == "c"
